/*****************************************************************************
aft_tdm_dtmf.c: Software DTMF sender and receiver utility
*
* Author(s):    David Yat Sin <davidy@sangoma.com>
*
* Copyright:	(c) 1995-2008 Sangoma Technologies Inc.
*
*		This program is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
* ============================================================================
* Description:
*
* 	The aft_tdm_dtmf utility will bind to a socket to listen to incoming 
* 	DTMF tones
* 	
*/



#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <if_wanpipe.h>
#include <string.h>
#include <signal.h>
#include <wait.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <pthread.h>
#include <wanpipe.h>
#include <wanpipe_tdm_api.h>


#include <libsangoma.h>

#include <libteletone.h>
#include <g711.h>
#include <aft_tdm_switch_buffer.h>


#define MAX_NUM_OF_TIMESLOTS  31*16

#define MAX_RX_DATA	5000
#define MAX_TX_DATA     5000

#define MAX_IF_NAME 20

#define u8 	unsigned char
#define u16	unsigned short
#define s16	signed short

#define SMG_DTMF_ON 	220
#define SMG_DTMF_OFF 	10
#define SMG_DTMF_RATE  	8000

/* For FXO hook-state control */

static unsigned char main_menu_list[]="\n"
"\n"
"A200: Main Menu\n" 
"-----------------\n"
"\n"
"0. Transmit On-hook\n"
"1. Transmit Off-hook\n"
"2. Transmit DTMF\n"
"q. Quit\n"
"\n"
"Please Select (<opt>|q): ";

typedef struct {
	int sock;
	int rx_cnt;
	int tx_cnt;
	int data;
	int last_error;
	int frames;
	char if_name[MAX_IF_NAME+1];
	int mtu;
	
	teletone_dtmf_detect_state_t 	dtmf_detect;
	teletone_generation_session_t 	tone_session;
	switch_buffer_t 		*dtmf_buffer;
	
	pthread_t 	tid_run;
	pthread_mutex_t	dtmf_lock;
	
} timeslot_t;


timeslot_t tslot_array[MAX_NUM_OF_TIMESLOTS];
int num_tslots;

int end=0;




void convert_ulaw_to_linear(u16 *linear_buffer, u8 *ulaw_buffer, int size)
{
	int i;
	memset(linear_buffer, 0, size);
	for(i = 0; i < size; i++) {
		linear_buffer[i] = ulaw_to_linear (ulaw_buffer[i]);
	}
}



void convert_linear_to_ulaw(u8 *ulaw_buffer, u16 *linear_buffer, int size)
{
	int i;
	memset(ulaw_buffer, 0, size);
	for(i = 0; i < size; i++) {
		ulaw_buffer[i] = linear_to_alaw (ulaw_buffer[i]);
	}
}


void print_packet(unsigned char *buf, int len)
{
	int nonzero = 0;
	int x;

	for(x = 0; x<len; x++) {
		if (buf[x] != 0) {
			nonzero = 1;
			break;
		}
	}
	if (nonzero == 0) {
		return;
	}	
	

	printf("{  | ");
	for (x=0;x<len;x++){
		if (x && x%24 == 0){
			printf("\n  ");
		}
		if (x && x%8 == 0)
			printf(" | ");
		printf("%02x ",buf[x]);
	}
	printf("}\n");
}

void sig_end(int sigid)
{
#if 0
	int i;
	wanpipe_tdm_api_t tdm_api;
#endif
	printf("%d: Got Signal %i\n",getpid(),sigid);
	end=1;

#if 0	
	printf("Setting FXO ON HOOK\n");
	for(i=1; i < num_tslots; i++) {
		if (sangoma_tdm_txsig_onhook(tslot_array[i].sock, &tdm_api)) {
			printf("FAILED TO SET  FXO ON HOOK\n");
		}
	}
#endif
}




/*===================================================
 * MakeConnection
 *
 *   o Create a Socket
 *   o Bind a socket to a wanpipe network interface
 *       (Interface name is supplied by the user)
 *==================================================*/         

int MakeConnection(timeslot_t *slot, char *router_name) 
{

	int span,chan;
	sangoma_span_chan_fromif(slot->if_name,&span,&chan);
	
	if (span > 0 && chan > 0) {
		wanpipe_tdm_api_t tdm_api;
             	slot->sock = sangoma_open_tdmapi_span_chan(span,chan);
	     	if( slot->sock < 0 ) {
      			perror("Open Span Chan: ");
      			return( FALSE );
		}

                sangoma_tdm_set_codec(slot->sock,&tdm_api,WP_NONE);	 
		slot->mtu = sangoma_tdm_get_usr_mtu_mru(slot->sock, &tdm_api);
		
		printf("Socket bound to Span=%i Chan=%i\n\n",span,chan);
		
#if 0
		printf("Setting FXO OFF HOOK\n");
		if (sangoma_tdm_txsig_offhook(slot->sock, &tdm_api)) {
			printf("FAILED TO SET  FXO OFF HOOK\n");
		}
#endif
		
		return (TRUE);
	}
	
	return(FALSE);
}

static int waitfor_tx_socket(int fd, int timeout, int flags)
{
	struct pollfd pfds[1];
	int res;
 
	memset(&pfds[0], 0, sizeof(pfds[0]));
	pfds[0].fd = fd;
	pfds[0].events = flags;
	res = poll(pfds, 1, timeout);

	if (res > 0) {
		if (pfds[0].revents & POLLOUT) {
			res = 1;
		} else if ((pfds[0].revents & POLLERR)) {
			res = -1;
		} else if ((pfds[0].revents & POLLNVAL)) {
			res = -2;
		} else {
			res = -1;
		}
	}

	return res;
}



int dtmf_transmit(timeslot_t *slot)
{
	
	int bread;
	
	u8 linear_data[MAX_TX_DATA];
	u8 ulaw_data[MAX_TX_DATA];
	
	
	//unsigned char dtmf[1024];
	//unsigned char dtmf_law[1024];
	sangoma_api_hdr_t hdrframe;
	int i;
	int slin_len = (slot->mtu)*2;
	
	short *data;
	int used;
	int res;
	int err;
	
	memset(&hdrframe,0,sizeof(hdrframe));
	for (;;) {
		
		if ((used=switch_buffer_inuse(slot->dtmf_buffer)) <= 0) {
			break;
		}

		res = waitfor_tx_socket(slot->sock, -1,  POLLERR | POLLOUT);
		if (res <= 0) {
			break;
		}

	//	pthread_mutex_lock(&slot->dtmf_lock);
		if ((used=switch_buffer_inuse(slot->dtmf_buffer)) <= 0) {
		//	pthread_mutex_unlock(&slot->dtmf_lock);
			break;
		}

		bread = switch_buffer_read(slot->dtmf_buffer, linear_data, slin_len);
	//	pthread_mutex_unlock(&slot->dtmf_lock);

		if (bread <= 0) {
			break;
		}
		
		data=(short*)linear_data;
		for (i=0;i<slot->mtu;i++) {
			ulaw_data[i] = linear_to_ulaw((int)data[i]);	
		}
	
		err=sangoma_sendmsg_socket(slot->sock,
					   &hdrframe, 
					   sizeof(hdrframe), 
	       			           ulaw_data, slot->mtu, 0);	
		if (err != slot->mtu) {
			printf("Error: Failed to TX to TDM API on DTMF (err=%i mtu=%i)!\n",err,slot->mtu);
		}

		

	} 

	return 0;
}


void *listen_dtmf(void* slot) 
{
	timeslot_t* tslot;
	tslot = slot;
	
	int err, dtmf_digit;
	
	char digit_str[100] = "";
	
	wp_tdm_api_rx_element_t* api_rx_el;
	fd_set 	ready;
	
	u16 linear_data[MAX_RX_DATA];
	u8  ulaw_data[MAX_RX_DATA];
	
	
	for(;;) {	
		//sleep(1);
		FD_ZERO(&ready);
		FD_SET(tslot->sock, &ready);

		if(select((tslot->sock+1),&ready, NULL, NULL, NULL)){
			if (FD_ISSET(tslot->sock,&ready)){
				memset(ulaw_data, 0, sizeof(ulaw_data));
				
				err = sangoma_readmsg_tdm(tslot->sock, 
							   ulaw_data, 
      							   sizeof(wp_tdm_api_rx_hdr_t), 
							   &ulaw_data[sizeof(wp_tdm_api_rx_hdr_t)],
							   MAX_RX_DATA, 0);
				
				if (err <= 0) {
					printf("%s: Error receiving data\n", tslot->if_name);
					break;
				} 
#if 0
				else {
					printf("%s: data %d\n", tslot->if_name, err);
				}
#endif
		
				api_rx_el = (wp_tdm_api_rx_element_t*)&ulaw_data[0];
				
				convert_ulaw_to_linear(&linear_data[0], &ulaw_data[sizeof(wp_tdm_api_rx_hdr_t)], err);
				//print_packet(linear_data, err);
				//print_packet(&linear_data[0], err);
			
				dtmf_digit = teletone_dtmf_detect(&tslot->dtmf_detect, (s16*) linear_data ,err);
				teletone_dtmf_get(&tslot->dtmf_detect, digit_str, sizeof(digit_str));

				if(*digit_str) {
					printf("%s: DTMF received %s\n", tslot->if_name, digit_str);
				} 				
			}
		}
	}
	
	pthread_exit(NULL);
} 


void *play_dtmf(void* ptr)
{
	char* digits;
	char *cur = NULL;
	int err;
	int dtmf_size;
	int wrote;
	int x;
	int tslot_index;

	dtmf_size=(SMG_DTMF_ON+SMG_DTMF_OFF)*10*2;

	digits = malloc(30);
	if (digits == NULL) {
		printf("Failed to allocate digits\n");
		exit(1);
	}
		
	for(;;) {
tslot_select:
		printf("Select timeslot to play DTMF tones:\n");
		for(x = 1; x < num_tslots; x++) {
			printf("%d . %s\n", x, tslot_array[x].if_name);
		}
		printf("Timeslot:");
		scanf("%d", &tslot_index);
		printf("\n");
		if(tslot_index >= num_tslots) {
			printf("Invalid tslot:%d\n\n", tslot_index);
			goto tslot_select;
		}
		
		if (!tslot_array[tslot_index].dtmf_buffer) {
			printf("%s: allocating DTMF buffer...", tslot_array[tslot_index].if_name);
			err=switch_buffer_create_dynamic(&tslot_array[tslot_index].dtmf_buffer, 1024, dtmf_size, 0);
			if (err != 0) {
				printf("Failed\n\n");
				return NULL;
			} else {
				printf("OK\n\n");
			}
		}
		
		for(;;) {
			printf("Input DTMF:\n");
			scanf("%s", digits);
			printf("\n\n");
			
			for (cur = digits; *cur; cur++) {			
				if ((wrote = teletone_mux_tones(&tslot_array[tslot_index].tone_session,
				     &tslot_array[tslot_index].tone_session.TONES[(int)*cur]))) {
				     
					pthread_mutex_lock(&tslot_array[tslot_index].dtmf_lock);
				
					err=switch_buffer_write(tslot_array[tslot_index].dtmf_buffer,
								tslot_array[tslot_index].tone_session.buffer, 
								wrote);
					
					pthread_mutex_unlock(&tslot_array[tslot_index].dtmf_lock);
				
					printf("%s: queuing DTMF %c Wrote=%i (sent=%i)\n",
								    tslot_array[tslot_index].if_name, *cur,wrote*2,err);
				     } else {
					     printf("Error: Sending DTMF %d (err=%i)\n",
							     *cur,wrote);
				     }
			
			}
			dtmf_transmit(&tslot_array[tslot_index]);		
		}
		
		
		
	}
	pthread_exit(NULL);
}


int tx_dtmf(int tslot_index )
{
	int digits;
	int wrote,err;
	char *cur = NULL;

	digits = malloc(30);
		
		for(;;) {
			printf("Input DTMF to transmit:\n");
			scanf("%s", digits);
			printf("\n\n");
			
			for (cur = digits; *cur; cur++) {			
				if ((wrote = teletone_mux_tones(&tslot_array[tslot_index].tone_session,
				     &tslot_array[tslot_index].tone_session.TONES[(int)*cur]))) {
				     
					//pthread_mutex_lock(&tslot_array[tslot_index].dtmf_lock);
				
					err=switch_buffer_write(tslot_array[tslot_index].dtmf_buffer,
								tslot_array[tslot_index].tone_session.buffer, 
								wrote);
					
				//	pthread_mutex_unlock(&tslot_array[tslot_index].dtmf_lock);
				
					printf("%s: queuing DTMF %c Wrote=%i (sent=%i)\n",
								    tslot_array[tslot_index].if_name, *cur,wrote*2,err);
				     } else {
					     printf("Error: Sending DTMF %d (err=%i)\n",
							     *cur,wrote);
				     }
			
			}
			dtmf_transmit(&tslot_array[tslot_index]);
			return 0;		
		}

}
int go_off_hook( int socket, wanpipe_tdm_api_t *tdm_api){
	/*Tranmsit off hoook*/
	//sangoma_tdm_enable_rxhook_events(socket, &tdm_api);
	return sangoma_tdm_txsig_offhook(socket, tdm_api);
			

}

int go_on_hook( int socket, wanpipe_tdm_api_t *tdm_api){
	/*Tranmsit off hoook*/
	//sangoma_tdm_enable_rxhook_events(socket, &tdm_api);
	return sangoma_tdm_txsig_onhook(socket, tdm_api);
			
}

void main_menu(void)
{
	int val;
    unsigned int data;
    int err;
	int i=0;
	wanpipe_tdm_api_t tdm_api;
	memset(&tdm_api,0,sizeof(tdm_api));
	system("clear");
	for(;;){
		printf("Please Enter the port No\n");
		scanf("%d",&i);
		if(i>num_tslots){
		 printf(" Wrong Port No. Please re-Enter the Port No\n");
		 }else{
			break;
		}
		
		
	}
	for(;;){

		val=-1;
      	printf("%s",main_menu_list);
      	scanf("%d",&val);
      	if (val==-1){
        	    getchar();
      	}
   
      	printf("Val=%d\n",val);
		
/*
0. Transmit On-hook\n"
"1. Transmit Off-hook\n"
"2. Transmit DTMF\n"
"q. Quit\n"
*/

		switch (val){
		case 0:
			/* Transmit On-hook */
			err=go_on_hook(tslot_array[i].sock, &tdm_api);
			if(err){
				printf("Failed to go FXO %d on-hook\n",i+1);
			}
			break;
	
		case 1:
			/* Transmit Off-hook */
			err=go_off_hook(tslot_array[i].sock, &tdm_api);
			if(err){
				printf("Failed to go FXO %d off-hook\n",i+1);
			}
			break;
	
		case 2:
			/* Transmit DTMF */
			err=tx_dtmf(i);
			break;
		default:
			
			end=1;
			printf("Exiting..\n");

		return;
	
		}  
	}



}

/***************************************************************
 * Main:
 *
 *    o Make a socket connection to the driver.
 *    o Call process_con() to read/write the socket 
 *
 **************************************************************/  

int main(int argc, char* argv[])
{
	int proceed;
	char router_name[20];	
	pthread_t tid_play_dtmf;
	int x,i,err;
	int menu_driven=1;
	int dtmf_size;
	dtmf_size=(SMG_DTMF_ON+SMG_DTMF_OFF)*10*2;
	if (argc < 3){
		printf("Usage: rec_wan_sock <router name> <interface name> <data> ...\n");
		exit(0);
	}	
	
	nice(-11);
	
	signal(SIGINT,&sig_end);
	signal(SIGTERM,&sig_end);
	
	memset(&tslot_array,0,sizeof(tslot_array));
	for (i=0;i<MAX_NUM_OF_TIMESLOTS;i++){
		tslot_array[i].sock=-1;
	}
	
	strncpy(router_name,argv[1],(sizeof(router_name)-1));	
	
	num_tslots = argc-1;
	printf("Number of num_tslots = %d , argc \%d \n",num_tslots,argc);
	
	/* Launch DTMF Listening threads */
	for (x=1;x<argc-1;x++) {	
		strncpy(tslot_array[x].if_name, argv[x+1], MAX_IF_NAME);		
		printf("Connecting to IF=%s %d\n",tslot_array[x].if_name,x);
	
		proceed = MakeConnection(&tslot_array[x],router_name);
		if( proceed == TRUE ){
			memset(&(tslot_array[x].tone_session), 0, sizeof(teletone_generation_session_t));
			/*JP buffer for dtmf */
			printf("%s: allocating DTMF buffer...", tslot_array[x].if_name);
			err=switch_buffer_create_dynamic(&tslot_array[x].dtmf_buffer, 1024, dtmf_size, 0);
			if (err != 0) {
				printf("Failed\n\n");
				return NULL;
			} else {
				printf("OK\n\n");
			}

			if (teletone_init_session(&(tslot_array[x].tone_session), 0, NULL, NULL)) {
				printf("%s :ERROR: Failed to initialize TONE!\n", tslot_array[x].if_name);			
			} else {
				tslot_array[x].tone_session.rate = SMG_DTMF_RATE;
				tslot_array[x].tone_session.duration = SMG_DTMF_ON * (SMG_DTMF_RATE / 1000);
				tslot_array[x].tone_session.wait = SMG_DTMF_OFF * (SMG_DTMF_RATE / 1000);
				//teletone_dtmf_detect_init (&(tslot_array[x].dtmf_detect), 8000);	
				//pthread_create(&(tslot_array[x].tid_run), NULL, listen_dtmf, (void*) &tslot_array[x]);
			}
		}else{
			if (tslot_array[x].sock){
				close(tslot_array[x].sock);
			}
			tslot_array[x].sock=-1;
		}
	}
	if (menu_driven){
		main_menu();
		//close(sock);
		//return 0;
	}else{
		/* Launch DTMF Generator threads */
		pthread_create(&tid_play_dtmf, NULL, play_dtmf, NULL);
	}
	
	/* Launch DTMF Generator threads */
//	pthread_create(&tid_play_dtmf, NULL, play_dtmf, NULL);
	
	
	while(end == 0) {
		sleep(5);
	}
	
	
	
	for (x=1;x<argc-1;x++) {
		if (tslot_array[x].sock){
			close(tslot_array[x].sock);
		}
	}
	tslot_array[x].sock=-1;
	
	
	return 0;
};
